export declare type ErrorMsg = string | Record<string, any>;
export declare type Context = {
    key?: string;
    index?: number;
};
export declare type Decoder = (value: any, context?: Context) => any;
export interface Switch<T> {
    decoder: T;
}
export declare type Schema = Record<string, Decoder | Switch<any>>;
export declare type TypeOf<T extends Decoder> = ReturnType<T>;
export declare type FilterOptionalKeys<T> = {
    [K in keyof T]: T[K] extends Switch<any> ? K : never;
}[keyof T];
export declare type FilterRequiredKeys<T> = {
    [K in keyof T]: T[K] extends Switch<any> ? never : K;
}[keyof T];
export declare type MergeIntersection<T extends Schema> = {
    [K in FilterOptionalKeys<T>]?: T[K] extends Switch<any> ? ReturnType<T[K]['decoder']> : never;
} & {
    [K in FilterRequiredKeys<T>]: ReturnType<T[K]>;
};
export declare type PartialMergeIntersection<T extends Schema> = {
    [K in FilterOptionalKeys<T>]: T[K] extends Switch<any> ? ReturnType<T[K]['decoder']> : never;
} & {
    [K in FilterRequiredKeys<T>]?: ReturnType<T[K]>;
};
export interface ObjectType {
    [index: string]: any;
}
export declare type DecoderValidator<T> = (val: T, context?: Context) => string | Record<string, string> | undefined;
export interface DecoderFullOption<T> {
    parse?(val: any): T;
    validate?: DecoderValidator<T> | DecoderValidator<T>[];
    errorMsg?: string;
    getErrorMsg?: (val: any, context?: Context) => string;
}
export declare type CustomErrorMsg = string;
export declare type DecoderOption<T> = DecoderValidator<T> | DecoderValidator<T>[] | CustomErrorMsg | DecoderFullOption<T>;
export declare function pipe<T>(validators: DecoderValidator<T>[]): (val: T, context?: Context | undefined) => string | Record<string, string> | undefined;
export declare function isObject(val: any): val is ObjectType;
export declare function isUndefined(val: any): val is undefined;
export declare class ValidationError extends Error {
    readonly error: ErrorMsg;
    constructor(msg: ErrorMsg);
}
export declare function toggle<T extends Decoder>(aDecoder: T): Switch<T>;
export declare function identity<T>(val: T): T;
interface DecoderParams<T, U> {
    typeGuard: (rawVal: unknown) => rawVal is T;
    getDefaultErrorMsg: (val: any, context?: Context) => string;
    defaultParser: (rawVal: U) => T | U;
}
export declare function decoder<T, U>({ typeGuard, getDefaultErrorMsg, defaultParser, }: DecoderParams<T, U>): (options?: DecoderOption<T>) => (val: any, context?: Context | undefined) => T;
export declare const string: (options?: DecoderOption<string>) => (val: any, context?: Context | undefined) => string;
export declare const number: (options?: DecoderOption<number>) => (val: any, context?: Context | undefined) => number;
export declare const boolean: (options?: DecoderOption<boolean>) => (val: any, context?: Context | undefined) => boolean;
export declare function optional(aDecoder: Decoder): <T>(val: any, context?: Context | undefined) => T | undefined;
export declare function getUnknownFields(inputKeys: string[], schemaKeys: string[]): string[];
export declare function constructUnknownFieldsError(unknownKeys: string[], unknownFieldErrorMsg: string): {};
interface ObjectFullOption<T> {
    strict?: boolean;
    unknownFieldErrorMsg?: string;
    validate?: DecoderValidator<T> | DecoderValidator<T>[];
}
declare type ObjectOption<T> = DecoderValidator<T> | ObjectFullOption<T>;
export declare function object<T extends Schema, U extends MergeIntersection<T>>(schema: T, option?: ObjectOption<{
    [K in keyof U]: U[K];
}>): (rawData: any, context?: Context | undefined) => { [K in keyof U]: U[K]; };
export declare function partial<T extends Schema, U extends PartialMergeIntersection<T>>(schema: T, customValidator?: DecoderValidator<{
    [K in keyof U]: U[K];
}>): (val: any, context?: Context | undefined) => { [K in keyof U]: U[K]; };
export declare function array<T extends Decoder>(schema: T, customValidator?: DecoderValidator<ReturnType<T>[]> | DecoderValidator<ReturnType<T>[]>[]): (data: any, context?: Context | undefined) => ReturnType<T>[];
export {};
