"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = function (d, b) {
        extendStatics = Object.setPrototypeOf ||
            ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
            function (d, b) { for (var p in b) if (Object.prototype.hasOwnProperty.call(b, p)) d[p] = b[p]; };
        return extendStatics(d, b);
    };
    return function (d, b) {
        if (typeof b !== "function" && b !== null)
            throw new TypeError("Class extends value " + String(b) + " is not a constructor or null");
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.array = exports.partial = exports.object = exports.constructUnknownFieldsError = exports.getUnknownFields = exports.optional = exports.boolean = exports.number = exports.string = exports.decoder = exports.identity = exports.toggle = exports.ValidationError = exports.isUndefined = exports.isObject = exports.pipe = void 0;
function pipe(validators) {
    return function (val, context) {
        for (var _i = 0, validators_1 = validators; _i < validators_1.length; _i++) {
            var v = validators_1[_i];
            var error = v(val, context);
            if (error !== undefined)
                return error;
        }
    };
}
exports.pipe = pipe;
function isObject(val) {
    return typeof val === 'object' && val !== null;
}
exports.isObject = isObject;
function isUndefined(val) {
    return typeof val === 'undefined';
}
exports.isUndefined = isUndefined;
var ValidationError = /** @class */ (function (_super) {
    __extends(ValidationError, _super);
    function ValidationError(msg) {
        var _this = _super.call(this, JSON.stringify(msg)) || this;
        _this.error = msg;
        return _this;
    }
    return ValidationError;
}(Error));
exports.ValidationError = ValidationError;
function toggle(aDecoder) {
    return {
        decoder: aDecoder,
    };
}
exports.toggle = toggle;
function identity(val) {
    return val;
}
exports.identity = identity;
function isDecoderValidator(decoderOption) {
    return typeof decoderOption === 'function';
}
function isCustomErrorMsg(decoderOption) {
    return typeof decoderOption === 'string';
}
function decoder(_a) {
    var typeGuard = _a.typeGuard, getDefaultErrorMsg = _a.getDefaultErrorMsg, defaultParser = _a.defaultParser;
    return function (options) {
        if (options === void 0) { options = {}; }
        return function (val, context) {
            val = defaultParser(val);
            options = isDecoderValidator(options)
                ? { validate: pipe([options]) }
                : options;
            options = Array.isArray(options)
                ? { validate: pipe(options) }
                : options;
            options = isCustomErrorMsg(options)
                ? { errorMsg: options }
                : options;
            if (Array.isArray(options.validate)) {
                options.validate = pipe(options.validate);
            }
            var parse = options.parse, validate = options.validate, errorMsg = options.errorMsg, getErrorMsg = options.getErrorMsg;
            val = parse ? parse(val) : val;
            if (!typeGuard(val))
                throw new ValidationError(errorMsg ||
                    (getErrorMsg && getErrorMsg(val, context)) ||
                    getDefaultErrorMsg(val, context));
            if (validate) {
                var error = validate(val, context);
                if (error)
                    throw new ValidationError(error);
            }
            return val;
        };
    };
}
exports.decoder = decoder;
exports.string = decoder({
    typeGuard: function (val) { return typeof val === 'string'; },
    getDefaultErrorMsg: function () { return 'Not a string'; },
    defaultParser: identity,
});
exports.number = decoder({
    typeGuard: function (val) { return typeof val === 'number'; },
    getDefaultErrorMsg: function () { return 'Not a number'; },
    defaultParser: identity,
});
exports.boolean = decoder({
    typeGuard: function (val) { return typeof val === 'boolean'; },
    getDefaultErrorMsg: function () { return 'Not a boolean'; },
    defaultParser: identity,
});
function isDecoder(v) {
    return typeof v === 'function';
}
function optional(aDecoder) {
    return function (val, context) {
        if (val === undefined)
            return undefined;
        return aDecoder(val, context);
    };
}
exports.optional = optional;
function mapValueError(data) {
    return function (_a) {
        var key = _a[0], aDecoder = _a[1];
        try {
            var actualDecoder = isDecoder(aDecoder)
                ? aDecoder
                : optional(aDecoder.decoder);
            return {
                key: key,
                value: actualDecoder(data[key], { key: key }),
                error: undefined,
            };
        }
        catch (e) {
            if (e.error !== undefined) {
                return { key: key, value: undefined, error: e.error };
            }
            throw e;
        }
    };
}
function getUnknownFields(inputKeys, schemaKeys) {
    return inputKeys.filter(function (key) { return !schemaKeys.includes(key); });
}
exports.getUnknownFields = getUnknownFields;
function constructUnknownFieldsError(unknownKeys, unknownFieldErrorMsg) {
    return unknownKeys.reduce(function (a, k) {
        var _a;
        return (__assign(__assign({}, a), (_a = {}, _a[k] = unknownFieldErrorMsg, _a)));
    }, {});
}
exports.constructUnknownFieldsError = constructUnknownFieldsError;
var defaultOption = { strict: true, unknownFieldErrorMsg: 'Unknown field.' };
function object(schema, option) {
    return function (rawData, context) {
        rawData = isObject(rawData) ? rawData : {};
        // convert option to ObjectFullOption
        option = isDecoderValidator(option) ? { validate: option } : option;
        option = isUndefined(option)
            ? defaultOption
            : __assign(__assign({}, defaultOption), option);
        if (Array.isArray(option.validate)) {
            option.validate = pipe(option.validate);
        }
        var _a = option, validate = _a.validate, strict = _a.strict, unknownFieldErrorMsg = _a.unknownFieldErrorMsg;
        // handle unknown fields
        if (strict) {
            var unknownKeys = getUnknownFields(Object.keys(rawData), Object.keys(schema));
            if (unknownKeys.length !== 0) {
                throw new ValidationError(constructUnknownFieldsError(unknownKeys, unknownFieldErrorMsg));
            }
        }
        // handle known fields
        var _b = Object.entries(schema)
            .map(mapValueError(rawData))
            .reduce(function (a, valueError) {
            var _a, _b;
            var key = valueError.key, value = valueError.value, anError = valueError.error;
            if (anError !== undefined) {
                return __assign(__assign({}, a), { error: __assign(__assign({}, a.error), (_a = {}, _a[key] = anError, _a)) });
            }
            return __assign(__assign({}, a), { data: __assign(__assign({}, a.data), (_b = {}, _b[key] = value, _b)) });
        }, { data: {}, error: {} }), data = _b.data, error = _b.error;
        if (Object.keys(error).length !== 0)
            throw new ValidationError(error);
        if (validate) {
            var err = validate(data, context);
            if (err)
                throw new ValidationError(err);
        }
        return data;
    };
}
exports.object = object;
function partial(schema, customValidator) {
    schema = Object.keys(schema).reduce(function (a, k) {
        var _a;
        var aDecoder = schema[k];
        aDecoder = isDecoder(aDecoder) ? optional(aDecoder) : aDecoder.decoder;
        return __assign(__assign({}, a), (_a = {}, _a[k] = aDecoder, _a));
    }, {});
    return function (val, context) {
        return object(schema, customValidator)(val, context);
    };
}
exports.partial = partial;
function array(schema, customValidator) {
    return function (data, context) {
        data = Array.isArray(data) ? data : [undefined];
        data = data.length === 0 ? [undefined] : data;
        var processedData = [];
        var errors = [];
        data.forEach(function (datum, index) {
            try {
                processedData.push(schema(datum, { index: index }));
            }
            catch (e) {
                if (e.error !== undefined) {
                    errors.push({ _index: index, error: e.error });
                }
                else {
                    throw e;
                }
            }
        });
        if (errors.length !== 0) {
            throw new ValidationError(errors);
        }
        if (Array.isArray(customValidator)) {
            customValidator = pipe(customValidator);
        }
        if (customValidator) {
            var err = customValidator(data, context);
            if (err)
                throw new ValidationError(err);
        }
        return data;
    };
}
exports.array = array;
